/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.tinydb.operator.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.springframework.dao.EmptyResultDataAccessException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.tinygroup.tinydb.Bean;
import org.tinygroup.tinydb.config.ColumnConfiguration;
import org.tinygroup.tinydb.config.TableConfiguration;
import org.tinygroup.tinydb.exception.DBRuntimeException;
import org.tinygroup.tinydb.operator.DBOperator;
import org.tinygroup.tinydb.operator.DbSingleOperator;
import org.tinygroup.tinydb.relation.Relation;
import org.tinygroup.tinydb.util.TinyDBUtil;

public class BeanDBSingleOperator<KeyType> extends BeanDBBaseOperator implements
		DbSingleOperator<KeyType> {

	public BeanDBSingleOperator(JdbcTemplate jdbcTemplate) {
		super(jdbcTemplate);
	}

	/**
	 * 更新一个对象
	 * 
	 * @param bean
	 * @param conditionColumns
	 * @param tableName
	 * 
	 * @return
	 */
	private int updateBean(Bean bean, List<String> conditionColumns) {
		List<ColumnConfiguration> paramsKeys = new ArrayList<ColumnConfiguration>();
		String sql = getUpdateSql(bean.getType(), conditionColumns, paramsKeys);
		List<Integer> dataTypes = getDataTypes(paramsKeys);
		List<Object> params = getParam(bean, paramsKeys);
		return executeByList(sql.toString(), params, dataTypes);
	}

	private Bean queryBean(KeyType beanId) {
		String sql = getQuerySqlByKey();
		return (Bean) queryObject(sql, getBeanType(), getSchema(), beanId);
	}


	public Bean insert(Bean bean) {
		insertTopBean(bean);// 新增最上层记录
		processRelation(bean, relation, new InsertRelationCallBack());
		return bean;
	}

	/**
	 * 
	 * bean对象描述的数据库记录是否存在
	 * 
	 * @param subBean
	 * @return
	 */
	@SuppressWarnings({ "rawtypes", "unchecked" })
	protected boolean beanNotExist(Bean bean) {
		DBOperator operator = getManager().getDbOperator(bean.getType());
		String primaryValue = getPrimaryKeyValue(operator, bean);
		if (primaryValue != null) {
			try {
				Bean selectBean = operator.getBean(primaryValue);
				if (selectBean != null) {
					return false;
				}
			} catch (Exception e) {
				if (e instanceof EmptyResultDataAccessException) {// jdbctemplate查询不存在会抛出此异常，这里判断记录是否存在，因此需要捕捉异常
					return true;
				}
			}

		}
		return true;
	}

	protected void insertTopBean(Bean bean) {
		String insertSql=getInsertSql(bean.getType());
		List<Object> params=getInsertParams(bean);
		List<Integer> dataTypes=getDataTypes(bean);
		executeByList(insertSql, params, dataTypes);
	}

	public int update(Bean bean) {
		int record = updateTopBean(bean);
		processRelation(bean, relation, new UpdateRelationCallBack());
		return record;

	}

	protected Bean[] getRelationBeans(Bean bean, String collectionMode,
			String keyName) {
		String relationType = null;
		if (collectionMode.equals(Relation.LIST_MODE)) {
			relationType = keyName + "List";
		} else if (collectionMode.equals(Relation.SET_MODE)) {
			relationType = keyName + "Set";
		} else if (collectionMode.equals(Relation.ARRAY_MODE)) {
			relationType = keyName + "Array";

		}
		Object object = bean.getProperty(relationType);
		return toBeanArrayFromObject(object);
	}

	@SuppressWarnings({ "unchecked", "rawtypes" })
	private static Bean[] toBeanArrayFromObject(Object object) {
		Bean[] beans = new Bean[0];
		if (object == null) {
			return beans;
		}
		if (Collection.class.isAssignableFrom(object.getClass())) {
			return (Bean[]) ((Collection) object).toArray(beans);
		}
		if (object.getClass().isArray()) {
			return (Bean[]) object;
		}
		return beans;
	}

	/**
	 * 
	 * 更新最上层bean
	 * 
	 * @param bean
	 * @return
	 */
	protected int updateTopBean(Bean bean) {
		TableConfiguration table = TinyDBUtil.getTableConfigByBean(
				bean.getType(), getSchema());
		List<String> conditionColumns = new ArrayList<String>();
		conditionColumns.add(table.getPrimaryKey().getColumnName());
		int record = updateBean(bean, conditionColumns);
		return record;
	}

	public int delete(Bean bean) {
		int record = deleteTopBean(bean);
		processRelation(bean, relation, new DeleteRelationCallBack());
		return record;
	}

	protected int deleteTopBean(Bean bean) {
		List<Integer> dataTypes=getDataTypes(bean);
		List<Object> params=getParams(bean);
		List<String> conditionColumns=getColumnNames(bean);
		String sql = getDeleteSql(bean.getType(), conditionColumns);
		return executeByList(sql, params, dataTypes);
	}

	public Bean getBean(KeyType beanId) {
		Bean bean = queryBean(beanId);
		bean = processRelation(bean, relation, new QueryRelationCallBack());
		return bean;
	}

	protected Bean processRelation(Bean bean, Relation relation,
			RelationCallBack callBack) {
		if (relation != null) {
			if (relation.getType().equals(bean.getType())) {
				List<Relation> relations = relation.getRelations();
				if (relations != null) {
					if (relation.getMode().equals(Relation.MORE_TO_ONE)) {// n对1
						callBack.processMoreToOne(bean, relations);
					} else if (relation.getMode().equals(Relation.ONE_TO_MORE)) {
						callBack.processOneToMore(bean, relation, relations);
					}

				}
			} else {
				throw new DBRuntimeException("tinydb.typeOfInconsistency",
						relation.getType(), bean.getType());
			}
		}
		return bean;

	}

	protected void checkOneToMore(List<Relation> relations) {
		Set<String> set = new HashSet<String>();
		for (Relation relation : relations) {
			set.add(relation.getType());
		}
		if (set.size() != 1 && relations.size() > 0) {
			throw new DBRuntimeException("tinydb.oneToMoreError");
		}
	}

	protected void checkMoreToOne(List<Relation> relations) {
		Set<String> set = new HashSet<String>();
		for (Relation relation : relations) {
			set.add(relation.getType());
		}
		if (set.size() != relations.size()) {
			throw new DBRuntimeException("tinydb.moreToOneError");
		}
	}

	@SuppressWarnings("unchecked")
	protected Bean createRelationBean(Bean bean, Relation relation) {
		String relationType = relation.getType();
		DBOperator<KeyType> operator = (DBOperator<KeyType>) getManager()
				.getDbOperator(relationType);
		Bean queryBean = new Bean(relationType);
		queryBean.setProperty(relation.getKeyName(),
				bean.getProperty(relation.getMainName()));
		Bean[] beans = operator.getBeans(queryBean);
		if (beans != null && beans.length == 1) {
			return beans[0];
		}
		throw new DBRuntimeException("tinydb.onlyUniqueRecords");
	}

	@SuppressWarnings("unchecked")
	protected Bean[] createRelationBeans(Bean bean, Relation relation) {
		String relationType = relation.getType();
		DBOperator<KeyType> operator = (DBOperator<KeyType>) getManager()
				.getDbOperator(relationType);
		Bean subBean = new Bean(relationType);
		subBean.setProperty(relation.getKeyName(),
				bean.getProperty(relation.getMainName()));
		return operator.getBeans(subBean);
	}

	public int deleteById(KeyType beanId) {
		String sql = getDeleteSqlByKey(getBeanType());
		TableConfiguration table = TinyDBUtil.getTableConfigByBean(
				getBeanType(), getSchema());
		List<Object> params = new ArrayList<Object>();
		params.add(beanId);
		List<Integer> dataTypes = new ArrayList<Integer>();
		dataTypes.add(table.getPrimaryKey().getDataType());
		return executeByList(sql, params, dataTypes);
	}

	/**
	 * 
	 * 功能说明: 关联处理的回调接口
	 * <p>
	 * 系统版本: v1.0<br>
	 * 开发人员: renhui <br>
	 * 开发时间: 2013-8-8 <br>
	 * 功能描述: 写明作用，调用方式，使用场景，以及特殊情况<br>
	 */
	interface RelationCallBack {
		void processMoreToOne(Bean bean, List<Relation> relations);

		void processOneToMore(Bean bean, Relation relation,
				List<Relation> relations);
	}

	abstract class CrudRelationCallBack implements RelationCallBack {

		public void processMoreToOne(Bean bean, List<Relation> relations) {
			checkMoreToOne(relations);
			for (Relation subRelation : relations) {
				Bean subBean = bean.getProperty(subRelation.getType());
				subBean.setProperty(subRelation.getKeyName(),
						bean.getProperty(subRelation.getMainName()));
				if (subBean != null) {
					moreToOnecallBack(subBean);
					processRelation(subBean, subRelation, this);
				}
			}
		}

		public void processOneToMore(Bean bean, Relation relation,
				List<Relation> relations) {
			checkOneToMore(relations);
			for (Relation subRelation : relations) {
				Bean[] beans = getRelationBeans(bean,
						relation.getCollectionMode(), subRelation.getKeyName());
				if (beans.length > 0) {
					for (Bean subBean : beans) {
						subBean.setProperty(subRelation.getKeyName(),
								bean.getProperty(subRelation.getMainName()));
						oneToMorecallBack(subBean);
						processRelation(subBean, subRelation, this);
					}
				}
			}
		}

		abstract void moreToOnecallBack(Bean bean);

		abstract void oneToMorecallBack(Bean bean);

	}

	class InsertRelationCallBack extends CrudRelationCallBack {
		@Override
		void moreToOnecallBack(Bean bean) {
			if (beanNotExist(bean)) {
				insertTopBean(bean);
			}
		}

		@Override
		void oneToMorecallBack(Bean bean) {
			if (beanNotExist(bean)) {
				insertTopBean(bean);
			}
		}
	}

	class UpdateRelationCallBack extends CrudRelationCallBack {
		@Override
		void moreToOnecallBack(Bean bean) {
			if (beanNotExist(bean)) {
				insertTopBean(bean);
			} else {
				updateTopBean(bean);
			}

		}

		@Override
		void oneToMorecallBack(Bean bean) {
			if (beanNotExist(bean)) {
				insertTopBean(bean);
			} else {
				updateTopBean(bean);
			}

		}
	}

	class DeleteRelationCallBack extends CrudRelationCallBack {
		@Override
		void moreToOnecallBack(Bean bean) {
			if (!beanNotExist(bean)) {
				deleteTopBean(bean);
			}

		}

		@Override
		void oneToMorecallBack(Bean bean) {
			if (!beanNotExist(bean)) {
				deleteTopBean(bean);
			}

		}
	}

	class QueryRelationCallBack implements RelationCallBack {
		public void processMoreToOne(Bean bean, List<Relation> relations) {
			checkMoreToOne(relations);
			for (Relation subRelation : relations) {
				Bean subBean = createRelationBean(bean, subRelation);
				subBean = processRelation(subBean, subRelation, this);
				bean.setProperty(subRelation.getType(), subBean);// 把关联对象存储到外层bean对象中
			}

		}

		public void processOneToMore(Bean bean, Relation relation,
				List<Relation> relations) {
			checkOneToMore(relations);
			for (Relation subRelation : relations) {
				Bean[] beans = createRelationBeans(bean, subRelation);// 创建一对多关联的beans
				List<Bean> subBeans = new ArrayList<Bean>();
				if (beans != null) {
					subBeans = Arrays.asList(beans);
				}
				if (subBeans.size() > 0) {
					for (Bean subBean : subBeans) {
						processRelation(subBean, subRelation, this);
					}
					String keyName = subRelation.getKeyName();
					if (relation.getCollectionMode().equals(Relation.LIST_MODE)) {
						bean.setProperty(keyName + "List", subBeans);
					} else if (relation.getCollectionMode().equals(
							Relation.SET_MODE)) {
						bean.setProperty(keyName + "Set", new HashSet<Bean>(
								subBeans));
					} else if (relation.getCollectionMode().equals(
							Relation.ARRAY_MODE)) {
						bean.setProperty(keyName + "Array",
								subBeans.toArray(new Bean[0]));
					}
				}

			}

		}
	}

}
